# Simple example inheritance and ABC using Shape/Square/Circle defined in the notes.


import math   # use pi
import abc    # use the class ABC and the function abstractmethod (a decorater)

# Shape is 
class Shape(abc.ABC):
    @abc.abstractmethod
    def perimeter(self): pass  # will be overridden in some derived class

    @abc.abstractmethod
    def area(self): pass       # will be overridden in some derived class

    def one_dimensional(self): # concrete: works by calling an abstract method
        return self.area() == 0

# Concrete classes derived from Shape; besides an __init__ specifying their
#   attributes, they specify concrete perimeter/area methods, overriding the
#   abstract methods defined in (and inherited from) the Shape ABC
# We will learn later in these notes how to easily ensure side and radius are >0

class Square(Shape):
    def __init__(self,side)  : self.side = side
    def __repr__(self)       : return f'Square(side={self.side})'
    def perimeter(self)      : return 4*self.side
    def area(self)           : return self.side**2

class Circle(Shape):
    def __init__(self,radius): self.radius = radius
    def __repr__(self)       : return f'Circle(radius={self.radius})'
    def perimeter(self)      : return math.pi * 2*self.radius
    def area(self)           : return math.pi * self.radius**2


try:
    shape_try = Shape()
except:
    print('Exception raised: could not construct Shape object')

asquare = Square(2)
print(f'\n{asquare} perimieter={asquare.perimeter()} area={asquare.area()} 1d={asquare.one_dimensional()}')

acircle = Circle(2)
print(f'\n{acircle} perimieter={acircle.perimeter()} area={acircle.area()} 1d={acircle.one_dimensional()}')
