# Full definition of the Unique_description_base and Constrained ABCs
#  in an inheritance hierarchy from the notes.
# The concrete class Non_negative is derived from Constrained;see similar
#  classes defined in the notes: Correcting_non_negative and Type_constrained.
# Uses __set__name__ instead of the fix_descriptor_names decorator.


class Unique_descriptor_base:
    __master_id = 0
    
    def __init__(self):
        self.__id = Unique_descriptor_base.__master_id
        Unique_descriptor_base.__master_id += 1
        
    def __set_name__(self, owner, name):
        self.special_attr_name = f'{name}@{self.__class__.__name__}#{self.__id}'


    # obj is the instance object this __get__ was called on; if it is None,
    #   then __get__ was called on this class not an instance of this class
    def __get__(self, obj, objtype = None):
        if obj is None:                      
            return self
        else:
            return getattr(obj, self.special_attr_name) 
    
    def __set__(self, obj, value):
        setattr(obj, self.special_attr_name, value) 

    def __delete__(self,obj):
        delattr(obj, self.special_attr_name)



import abc
class Constrained(abc.ABC, Unique_descriptor_base):
    def __set__(self,obj, value):
        value = self.check_constraint(obj, value)
        setattr(obj, self.special_attr_name, value)

    @abc.abstractmethod
    # Check constraint and do nothing if satisfied; raise ValueError if not
    #   satisfied; may also return a modified value that meets the constraint,
    #   allowing implicit conversion  
    def check_constraint(self, value):
        pass



class Non_negative(Constrained):
    def check_constraint(self, obj, value):
        if value >= 0:
            return value
        else:
            raise ValueError(f'Non_negative.check_constraint: value({value}) not >= 0')



class C:
    def __init__(self,init):
        self.count = init
        self.count2 = init+1

    def inc(self):
        self.count += 1

    #...def other methods...

    count  = Non_negative()
    count2 = Non_negative()


o = C(5)
print('o = C(5)')
print('o namespace=',o.__dict__)
print('search for o.count=',o.count)

print('\no.inc()')
o.inc()
print('search for o.count=',o.count)

print('\no.count = 1')
o.count = 1
print('search for o.count=',o.count)

print('\ndel o.count')
del o.count
print('o namespace=',o.__dict__)

try:
    print("\nprint('search for o.count=',o.count)")
    print('search for o.count=',o.count) 
except:
    print('Exception raised: could not compute o.count')
